from enum import IntEnum


class HttpStatusCode(IntEnum):
    Processing = 102
    OK = 200
    Created = 201
    Accepted = 202
    NoContent = 204
    MovedPermanently = 301
    Found = 302
    SeeOther = 303
    NotModified = 304
    BadRequest = 400
    Unauthorized = 401
    Forbidden = 403
    NotFound = 404
    MethodNotAllowed = 405
    Conflict = 409
    UnprocessableEntity = 422
    TooEarly = 425
    InternalServerError = 500
    NotImplemented = 501
    BadGateway = 502
    ServiceUnavailable = 503
    GatewayTimeout = 504
    InsufficientStorage = 507



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#
import logging
from asyncio import TimeoutError
from enum import unique, StrEnum
from typing import Union, Optional, Any, Mapping

from aiohttp import ClientResponse, ClientSession, BasicAuth, ClientTimeout, hdrs
from aiohttp.web_exceptions import HTTPBadRequest, HTTPUnauthorized, HTTPForbidden, HTTPNotFound, \
    HTTPUnprocessableEntity, HTTPBadGateway, HTTPServiceUnavailable, HTTPGatewayTimeout, HTTPInternalServerError, \
    HTTPError
from async_tools import AsyncOnStop

from .digest_auth import DigestAuth
from .decoder import Decoder
from .decoder_register import DecoderRegister
from .header_utils import parse_content_type_header
from .http_status_codes import HttpStatusCode
from .mime_types import ContentType

logger = logging.getLogger(__name__)


class HttpClient(AsyncOnStop):
    @unique
    class Method(StrEnum):
        GET = 'GET'
        HEAD = 'HEAD'
        PUT = 'PUT'
        POST = 'POST'
        PATCH = 'PATCH'
        DELETE = 'DELETE'

    def __init__(self, session: Optional[ClientSession], default_timeout_sec: float = 100):
        self._session = session
        self._default_timeout_sec = ClientTimeout(default_timeout_sec)
        self.decoder_register = DecoderRegister()

    async def _on_stop(self) -> None:
        await self._session.close()

    async def get(self,
                  url: str,
                  *,
                  params: Optional[dict] = None,
                  headers: Optional[Mapping[str, str]] = None,
                  auth: Optional[BasicAuth] = None,
                  timeout: Optional[float] = None,
                  force_decoder: Decoder = None,
                  proxy_url: Optional[str] = None
                  ) -> Any:
        return await self.request(self.Method.GET, url, params=params, timeout=timeout, auth=auth,
                                  headers=headers, force_decoder=force_decoder, proxy_url=proxy_url)

    async def head(self,
                   url: str,
                   *,
                   params: Optional[dict] = None,
                   headers: Optional[Mapping[str, str]] = None,
                   auth: Optional[BasicAuth] = None,
                   timeout: Optional[float] = None,
                   force_decoder: Decoder = None,
                   proxy_url: Optional[str] = None
                   ) -> Any:
        return await self.request(self.Method.HEAD, url, params=params, timeout=timeout, auth=auth,
                                  headers=headers, force_decoder=force_decoder, proxy_url=proxy_url)

    async def put(self,
                  url: str,
                  *,
                  params: Optional[dict] = None,
                  json: Union[dict, list, None] = None,
                  data: Optional[bytes] = None,
                  headers: Optional[Mapping[str, str]] = None,
                  auth: Optional[BasicAuth] = None,
                  timeout: Optional[float] = None,
                  force_decoder: Decoder = None,
                  proxy_url: Optional[str] = None
                  ) -> Any:
        return await self.request(self.Method.PUT, url, params=params, json=json, data=data, timeout=timeout,
                                  auth=auth, headers=headers, force_decoder=force_decoder, proxy_url=proxy_url)

    async def post(self,
                   url: str,
                   *,
                   params: Optional[dict] = None,
                   json: Union[dict, list, None] = None,
                   data: Optional[bytes] = None,
                   headers: Optional[Mapping[str, str]] = None,
                   auth: Optional[BasicAuth] = None,
                   timeout: Optional[float] = None,
                   force_decoder: Decoder = None,
                   proxy_url: Optional[str] = None
                   ) -> Any:
        return await self.request(self.Method.POST, url, params=params, json=json, data=data, timeout=timeout,
                                  auth=auth, headers=headers, force_decoder=force_decoder, proxy_url=proxy_url)

    async def patch(self,
                    url: str,
                    *,
                    params: Optional[dict] = None,
                    json: Union[dict, list, None] = None,
                    data: Optional[bytes] = None,
                    headers: Optional[Mapping[str, str]] = None,
                    auth: Optional[BasicAuth] = None,
                    timeout: Optional[float] = None,
                    force_decoder: Decoder = None,
                    proxy_url: Optional[str] = None
                    ) -> Any:
        return await self.request(self.Method.PATCH, url, params=params, json=json, data=data, timeout=timeout,
                                  auth=auth, headers=headers, force_decoder=force_decoder, proxy_url=proxy_url)

    async def delete(self,
                     url: str,
                     *,
                     params: Optional[dict] = None,
                     json: Union[dict, list, None] = None,
                     data: Optional[bytes] = None,
                     headers: Optional[Mapping[str, str]] = None,
                     auth: Optional[BasicAuth] = None,
                     timeout: Optional[float] = None,
                     force_decoder: Decoder = None,
                     proxy_url: Optional[str] = None
                     ) -> Any:
        return await self.request(self.Method.DELETE, url, params=params, json=json, data=data, timeout=timeout,
                                  auth=auth, headers=headers, force_decoder=force_decoder, proxy_url=proxy_url)

    async def request(self,
                      method: Method,
                      url: str,
                      params: Optional[dict] = None,
                      json: Union[dict, list, None] = None,
                      data: Optional[bytes] = None,
                      headers: Optional[Mapping[str, str]] = None,
                      auth: Optional[BasicAuth] = None,
                      timeout: Optional[float] = None,
                      force_decoder: Decoder = None,
                      proxy_url: Optional[str] = None
                      ) -> Any:
        client_timeout = ClientTimeout(timeout) if timeout else self._default_timeout_sec
        try:
            async with self.session.request(method, url, params=params, json=json, data=data, timeout=client_timeout,
                                            auth=auth, headers=headers, proxy=proxy_url) as response:
                return await self._process_response(response, force_decoder)
        except TimeoutError as er:
            logger.error(f"{method} {url} request exceeded the deadline = {client_timeout.total}. URL: {repr(er)}")
            raise er
        except HTTPUnauthorized as er:
            if not isinstance(auth, DigestAuth) or not auth.is_first_request:
                raise er
            auth.set_response_context(response)
            return await self.request(method, url, params, json, data, headers, auth, timeout, force_decoder)
        except HTTPError as er:
            logger.warning(f"{method} {url} request failed. Description: {repr(er)}")
            raise er

    @property
    def session(self) -> ClientSession:
        if self._session is None or self._session.closed:
            self._session = ClientSession()
        return self._session

    async def _process_response_payload(self, response: ClientResponse, force_decoder: Decoder = None) -> Any:
        payload = await response.read()
        if payload:
            content_type_header = response.headers.get(hdrs.CONTENT_TYPE, ContentType.Octet)
            content_type, content_type_info = parse_content_type_header(content_type_header)
            charset = content_type_info.get('charset')
            if charset:
                payload = payload.decode(charset)

            decoder = force_decoder if force_decoder is not None else self.decoder_register.pick_decoder(content_type)
            if decoder is not None:
                payload = decoder(payload)

        return payload

    async def _process_response(self, response: ClientResponse, force_decoder: Decoder = None) -> Any:
        if response.status in {HttpStatusCode.NoContent}:
            return

        payload = await self._process_response_payload(response, force_decoder)

        if response.status in {HttpStatusCode.OK, HttpStatusCode.Created, HttpStatusCode.Accepted}:
            return payload
        elif response.status == HttpStatusCode.BadRequest:
            raise HTTPBadRequest(reason=str(payload))
        elif response.status == HttpStatusCode.Unauthorized:
            raise HTTPUnauthorized(reason=str(payload))
        elif response.status == HttpStatusCode.Forbidden:
            raise HTTPForbidden(reason=str(payload))
        elif response.status == HttpStatusCode.NotFound:
            raise HTTPNotFound(reason=str(payload))
        elif response.status == HttpStatusCode.UnprocessableEntity:
            raise HTTPUnprocessableEntity(reason=str(payload))
        elif response.status == HttpStatusCode.BadGateway:
            raise HTTPBadGateway(reason=str(payload))
        elif response.status == HttpStatusCode.ServiceUnavailable:
            raise HTTPServiceUnavailable(reason=str(payload))
        elif response.status == HttpStatusCode.GatewayTimeout:
            raise HTTPGatewayTimeout(reason=str(payload))
        else:
            logger.warning(f"Not processed response status code = {response.status}. body: {payload}")
            raise HTTPInternalServerError(reason=str(payload))



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#
from typing import Optional, Mapping, Any

from init_helpers.custom_json import Jsonable

from .abstract import AbmServiceAuthedConnector


class AbmServicePayloadAuthConnector(AbmServiceAuthedConnector):
    async def get(self,
                  path: str,
                  url_query: Optional[Mapping[str, Any]] = None,
                  headers: Optional[Mapping[str, str]] = None,
                  token: str = None,
                  timeout_sec: Optional[float] = None) -> Any:
        raise RuntimeError("Attempt to make http GET request by payload auth connector, but GET request has no payload")

    async def head(self,
                   path: str,
                   url_query: Optional[Mapping[str, Any]] = None,
                   headers: Optional[Mapping[str, str]] = None,
                   token: str = None,
                   timeout_sec: Optional[float] = None) -> Any:
        raise RuntimeError(
            "Attempt to make http HEAD request by payload auth connector, but HEAD request has no payload")

    async def put(self,
                  path: str,
                  payload: bytes | Jsonable,
                  url_query: Optional[Mapping[str, Any]] = None,
                  headers: Optional[Mapping[str, str]] = None,
                  token: str = None,
                  timeout_sec: Optional[float] = None) -> Any:
        if isinstance(payload, bytes):
            raise NotImplementedError("Can not insert token into bytes")
        payload['token'] = token
        return await self._connector.put(
            path=path, payload=payload, url_query=url_query, headers=headers, timeout_sec=timeout_sec)

    async def post(self,
                   path: str,
                   payload: bytes | Jsonable,
                   url_query: Optional[Mapping[str, Any]] = None,
                   headers: Optional[Mapping[str, str]] = None,
                   token: str = None,
                   timeout_sec: Optional[float] = None) -> Any:
        if isinstance(payload, bytes):
            raise NotImplementedError("Can not insert token into bytes")
        payload['token'] = token
        return await self._connector.post(
            path=path, payload=payload, url_query=url_query, headers=headers, timeout_sec=timeout_sec)

    async def patch(self,
                    path: str,
                    payload: bytes | Jsonable,
                    url_query: Optional[Mapping[str, Any]] = None,
                    headers: Optional[Mapping[str, str]] = None,
                    token: str = None,
                    timeout_sec: Optional[float] = None) -> Any:
        if isinstance(payload, bytes):
            raise NotImplementedError("Can not insert token into bytes")
        payload['token'] = token
        return await self._connector.patch(
            path=path, payload=payload, url_query=url_query, headers=headers, timeout_sec=timeout_sec)

    async def delete(self,
                     path: str,
                     payload: bytes | Jsonable,
                     url_query: Optional[Mapping[str, Any]] = None,
                     headers: Optional[Mapping[str, str]] = None,
                     token: str = None,
                     timeout_sec: Optional[float] = None) -> Any:
        if isinstance(payload, bytes):
            raise NotImplementedError("Can not insert token into bytes")
        payload['token'] = token
        return await self._connector.delete(
            path=path, payload=payload, url_query=url_query, headers=headers, timeout_sec=timeout_sec)



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#
from typing import Tuple, Dict

from http_tools.mime_types import ContentType


def parse_content_type_header(content_type_header: str) -> Tuple[ContentType, Dict[str, str]]:
    content_type_header_parts = content_type_header.split(';')
    content_type = ContentType(content_type_header_parts.pop(0))
    content_type_info = {}
    for info in content_type_header_parts:
        info_parts = info.strip().split("=", 1)
        content_type_info[info_parts[0]] = info_parts[1] if len(info_parts) > 1 else ''
    return content_type, content_type_info



#  Copyright (C) 2021
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
#
from typing import Callable, Any

Decoder = Callable[[bytes], Any]



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mikhail Mamrov <m.mamrov@abm-jsc.ru>

import asyncio
import itertools
import logging
from functools import wraps
from time import sleep
from typing import Callable, Any, Union, Type, Optional


logger = logging.getLogger(__name__)


def retry(func: Callable[[Any], Any] = None,
          attempts_limit: Optional[int] = None,
          sleep_time_sec: float = 5.0,
          retrying_exceptions: Union[Type[Exception], tuple[Type[Exception], ...]] = Exception,
          ) -> Callable[[Any], Any]:
    def decorator(handler_) -> Callable[[Any], Any]:
        def handle_exception(er: Exception, attempt: int) -> None:
            logger.error(f"Execution '{handler_}' unsuccessful, because: {repr(er)}. Retry: {attempt}/{attempts_limit}")
            if attempts_limit is not None and attempt >= attempts_limit:
                raise

        if asyncio.iscoroutinefunction(handler_):
            @wraps(handler_)
            async def wrapper(*args, **kwargs) -> Callable[[Any], Any]:
                for attempt in itertools.count():
                    try:
                        return await handler_(*args, **kwargs)
                    except retrying_exceptions as er:
                        handle_exception(er, attempt)
                    await asyncio.sleep(sleep_time_sec)
            return wrapper
        else:
            @wraps(handler_)
            def wrapper(*args, **kwargs) -> Callable[[Any], Any]:
                for attempt in itertools.count():
                    try:
                        return handler_(*args, **kwargs)
                    except retrying_exceptions as er:
                        handle_exception(er, attempt)
                    sleep(sleep_time_sec)
            return wrapper

    return decorator if func is None else decorator(func)
